from __future__ import annotations

import datetime as dt
from copy import copy, deepcopy
from typing import NamedTuple

import pytest

from marshmallow import Schema, fields, utils


def test_missing_singleton_copy():
    assert copy(utils.missing) is utils.missing
    assert deepcopy(utils.missing) is utils.missing


class PointNT(NamedTuple):
    x: int | None
    y: int | None


class PointClass:
    def __init__(self, x, y):
        self.x = x
        self.y = y


class PointDict(dict):
    def __init__(self, x, y):
        super().__init__({"x": x})
        self.y = y


@pytest.mark.parametrize(
    "obj", [PointNT(24, 42), PointClass(24, 42), PointDict(24, 42), {"x": 24, "y": 42}]
)
def test_get_value_from_object(obj):
    assert utils.get_value(obj, "x") == 24
    assert utils.get_value(obj, "y") == 42


def test_get_value_from_namedtuple_with_default():
    p = PointNT(x=42, y=None)
    # Default is only returned if key is not found
    assert utils.get_value(p, "z", default=123) == 123
    # since 'y' is an attribute, None is returned instead of the default
    assert utils.get_value(p, "y", default=123) is None


class Triangle:
    def __init__(self, p1, p2, p3):
        self.p1 = p1
        self.p2 = p2
        self.p3 = p3
        self.points = [p1, p2, p3]


def test_get_value_for_nested_object():
    tri = Triangle(p1=PointClass(1, 2), p2=PointNT(3, 4), p3={"x": 5, "y": 6})
    assert utils.get_value(tri, "p1.x") == 1
    assert utils.get_value(tri, "p2.x") == 3
    assert utils.get_value(tri, "p3.x") == 5


# regression test for https://github.com/marshmallow-code/marshmallow/issues/62
def test_get_value_from_dict():
    d = dict(items=["foo", "bar"], keys=["baz", "quux"])
    assert utils.get_value(d, "items") == ["foo", "bar"]
    assert utils.get_value(d, "keys") == ["baz", "quux"]


def test_get_value():
    lst = [1, 2, 3]
    assert utils.get_value(lst, 1) == 2

    class MyInt(int):
        pass

    assert utils.get_value(lst, MyInt(1)) == 2


def test_set_value():
    d: dict[str, int | dict] = {}
    utils.set_value(d, "foo", 42)
    assert d == {"foo": 42}

    d = {}
    utils.set_value(d, "foo.bar", 42)
    assert d == {"foo": {"bar": 42}}

    d = {"foo": {}}
    utils.set_value(d, "foo.bar", 42)
    assert d == {"foo": {"bar": 42}}

    d = {"foo": 42}
    with pytest.raises(ValueError):
        utils.set_value(d, "foo.bar", 42)


def test_is_collection():
    assert utils.is_collection([1, "foo", {}]) is True
    assert utils.is_collection(("foo", 2.3)) is True
    assert utils.is_collection({"foo": "bar"}) is False


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (1676386740, dt.datetime(2023, 2, 14, 14, 59, 00)),
        (1676386740.58, dt.datetime(2023, 2, 14, 14, 59, 00, 580000)),
    ],
)
def test_from_timestamp(value, expected):
    result = utils.from_timestamp(value)
    assert type(result) is dt.datetime
    assert result == expected


def test_from_timestamp_with_negative_value():
    value = -10
    with pytest.raises(ValueError, match=r"Not a valid POSIX timestamp"):
        utils.from_timestamp(value)


def test_from_timestamp_with_overflow_value():
    value = 9223372036854775
    with pytest.raises(ValueError, match=r"out of range|year must be in 1\.\.9999"):
        utils.from_timestamp(value)


# Regression test for https://github.com/marshmallow-code/marshmallow/issues/540
def test_function_field_using_type_annotation():
    def get_split_words(value: str):
        return value.split(";")

    class MySchema(Schema):
        friends = fields.Function(deserialize=get_split_words)

    data = {"friends": "Clark;Alfred;Robin"}
    result = MySchema().load(data)
    assert result == {"friends": ["Clark", "Alfred", "Robin"]}



import datetime as dt

from marshmallow import EXCLUDE, Schema, fields


class UserSchema(Schema):
    name = fields.String(allow_none=True)
    email = fields.Email(allow_none=True)
    age = fields.Integer()
    created = fields.DateTime()
    id = fields.Integer(allow_none=True)
    homepage = fields.Url()
    birthdate = fields.Date()


class ProfileSchema(Schema):
    user = fields.Nested(UserSchema)


class TestFieldOrdering:
    def test_declared_field_order_is_maintained_on_dump(self, user):
        ser = UserSchema()
        data = ser.dump(user)
        keys = list(data)
        assert keys == [
            "name",
            "email",
            "age",
            "created",
            "id",
            "homepage",
            "birthdate",
        ]

    def test_declared_field_order_is_maintained_on_load(self, serialized_user):
        schema = UserSchema(unknown=EXCLUDE)
        data = schema.load(serialized_user)
        keys = list(data)
        assert keys == [
            "name",
            "email",
            "age",
            "created",
            "id",
            "homepage",
            "birthdate",
        ]

    def test_nested_field_order_with_only_arg_is_maintained_on_dump(self, user):
        schema = ProfileSchema()
        data = schema.dump({"user": user})
        user_data = data["user"]
        keys = list(user_data)
        assert keys == [
            "name",
            "email",
            "age",
            "created",
            "id",
            "homepage",
            "birthdate",
        ]

    def test_nested_field_order_with_only_arg_is_maintained_on_load(self):
        schema = ProfileSchema()
        data = schema.load(
            {
                "user": {
                    "name": "Foo",
                    "email": "Foo@bar.com",
                    "age": 42,
                    "created": dt.datetime.now().isoformat(),
                    "id": 123,
                    "homepage": "http://foo.com",
                    "birthdate": dt.datetime.now().date().isoformat(),
                }
            }
        )
        user_data = data["user"]
        keys = list(user_data)
        assert keys == [
            "name",
            "email",
            "age",
            "created",
            "id",
            "homepage",
            "birthdate",
        ]

    def test_nested_field_order_with_exclude_arg_is_maintained(self, user):
        class HasNestedExclude(Schema):
            user = fields.Nested(UserSchema, exclude=("birthdate",))

        ser = HasNestedExclude()
        data = ser.dump({"user": user})
        user_data = data["user"]
        keys = list(user_data)
        assert keys == ["name", "email", "age", "created", "id", "homepage"]


class TestIncludeOption:
    class AddFieldsSchema(Schema):
        name = fields.Str()

        class Meta:
            include = {"from": fields.Str()}

    def test_fields_are_added(self):
        s = self.AddFieldsSchema()
        in_data = {"name": "Steve", "from": "Oskosh"}
        result = s.load({"name": "Steve", "from": "Oskosh"})
        assert result == in_data

    def test_included_fields_ordered_after_declared_fields(self):
        class AddFieldsOrdered(Schema):
            name = fields.Str()
            email = fields.Str()

            class Meta:
                include = {
                    "from": fields.Str(),
                    "in": fields.Str(),
                    "@at": fields.Str(),
                }

        s = AddFieldsOrdered()
        in_data = {
            "name": "Steve",
            "from": "Oskosh",
            "email": "steve@steve.steve",
            "in": "VA",
            "@at": "Charlottesville",
        }
        # declared fields, then "included" fields
        expected_fields = ["name", "email", "from", "in", "@at"]
        assert list(AddFieldsOrdered._declared_fields.keys()) == expected_fields

        result = s.load(in_data)
        assert list(result.keys()) == expected_fields

    def test_added_fields_are_inherited(self):
        class AddFieldsChild(self.AddFieldsSchema):  # type: ignore[name-defined]
            email = fields.Str()

        s = AddFieldsChild()
        assert "email" in s._declared_fields
        assert "from" in s._declared_fields
        assert isinstance(s._declared_fields["from"], fields.Str)


class TestManyOption:
    class ManySchema(Schema):
        foo = fields.Str()

        class Meta:
            many = True

    def test_many_by_default(self):
        test = self.ManySchema()
        assert test.load([{"foo": "bar"}]) == [{"foo": "bar"}]

    def test_explicit_single(self):
        test = self.ManySchema(many=False)
        assert test.load({"foo": "bar"}) == {"foo": "bar"}



import typing

import pytest

from marshmallow import (
    Schema,
    fields,
    post_dump,
    post_load,
    pre_dump,
    pre_load,
    validates,
    validates_schema,
)
from marshmallow.exceptions import ValidationError
from marshmallow.experimental.context import Context
from tests.base import Blog, User


class UserContextSchema(Schema):
    is_owner = fields.Method("get_is_owner")
    is_collab = fields.Function(
        lambda user: user in Context[dict[str, typing.Any]].get()["blog"]
    )

    def get_is_owner(self, user):
        return Context.get()["blog"].user.name == user.name


class TestContext:
    def test_context_load_dump(self):
        class ContextField(fields.Integer):
            def _serialize(self, value, attr, obj, **kwargs):
                if (context := Context[dict].get(None)) is not None:
                    value *= context.get("factor", 1)
                return super()._serialize(value, attr, obj, **kwargs)

            def _deserialize(self, value, attr, data, **kwargs):
                val = super()._deserialize(value, attr, data, **kwargs)
                if (context := Context[dict].get(None)) is not None:
                    val *= context.get("factor", 1)
                return val

        class ContextSchema(Schema):
            ctx_fld = ContextField()

        ctx_schema = ContextSchema()

        assert ctx_schema.load({"ctx_fld": 1}) == {"ctx_fld": 1}
        assert ctx_schema.dump({"ctx_fld": 1}) == {"ctx_fld": 1}
        with Context({"factor": 2}):
            assert ctx_schema.load({"ctx_fld": 1}) == {"ctx_fld": 2}
            assert ctx_schema.dump({"ctx_fld": 1}) == {"ctx_fld": 2}

    def test_context_method(self):
        owner = User("Joe")
        blog = Blog(title="Joe Blog", user=owner)
        serializer = UserContextSchema()
        with Context({"blog": blog}):
            data = serializer.dump(owner)
            assert data["is_owner"] is True
            nonowner = User("Fred")
            data = serializer.dump(nonowner)
            assert data["is_owner"] is False

    def test_context_function(self):
        owner = User("Fred")
        blog = Blog("Killer Queen", user=owner)
        collab = User("Brian")
        blog.collaborators.append(collab)
        with Context({"blog": blog}):
            serializer = UserContextSchema()
            data = serializer.dump(collab)
            assert data["is_collab"] is True
            noncollab = User("Foo")
            data = serializer.dump(noncollab)
            assert data["is_collab"] is False

    def test_function_field_handles_bound_serializer(self):
        class SerializeA:
            def __call__(self, value):
                return "value"

        serialize = SerializeA()

        # only has a function field
        class UserFunctionContextSchema(Schema):
            is_collab = fields.Function(serialize)

        owner = User("Joe")
        serializer = UserFunctionContextSchema()
        data = serializer.dump(owner)
        assert data["is_collab"] == "value"

    def test_nested_fields_inherit_context(self):
        class InnerSchema(Schema):
            likes_bikes = fields.Function(lambda obj: "bikes" in Context.get()["info"])

        class CSchema(Schema):
            inner = fields.Nested(InnerSchema)

        ser = CSchema()
        with Context[dict]({"info": "i like bikes"}):
            obj: dict[str, dict] = {"inner": {}}
            result = ser.dump(obj)
            assert result["inner"]["likes_bikes"] is True

    # Regression test for https://github.com/marshmallow-code/marshmallow/issues/820
    def test_nested_list_fields_inherit_context(self):
        class InnerSchema(Schema):
            foo = fields.Raw()

            @validates("foo")
            def validate_foo(self, value, **kwargs):
                if "foo_context" not in Context[dict].get():
                    raise ValidationError("Missing context")

        class OuterSchema(Schema):
            bars = fields.List(fields.Nested(InnerSchema()))

        inner = InnerSchema()
        with Context({"foo_context": "foo"}):
            assert inner.load({"foo": 42})

        outer = OuterSchema()
        with Context({"foo_context": "foo"}):
            assert outer.load({"bars": [{"foo": 42}]})

    # Regression test for https://github.com/marshmallow-code/marshmallow/issues/820
    def test_nested_dict_fields_inherit_context(self):
        class InnerSchema(Schema):
            foo = fields.Raw()

            @validates("foo")
            def validate_foo(self, value, **kwargs):
                if "foo_context" not in Context[dict].get():
                    raise ValidationError("Missing context")

        class OuterSchema(Schema):
            bars = fields.Dict(values=fields.Nested(InnerSchema()))

        inner = InnerSchema()
        with Context({"foo_context": "foo"}):
            assert inner.load({"foo": 42})

        outer = OuterSchema()
        with Context({"foo_context": "foo"}):
            assert outer.load({"bars": {"test": {"foo": 42}}})

    # Regression test for https://github.com/marshmallow-code/marshmallow/issues/1404
    def test_nested_field_with_unpicklable_object_in_context(self):
        class Unpicklable:
            def __deepcopy__(self, _):
                raise NotImplementedError

        class InnerSchema(Schema):
            foo = fields.Raw()

        class OuterSchema(Schema):
            inner = fields.Nested(InnerSchema())

        outer = OuterSchema()
        obj = {"inner": {"foo": 42}}
        with Context({"unp": Unpicklable()}):
            assert outer.dump(obj)

    def test_function_field_passed_serialize_with_context(self, user):
        class Parent(Schema):
            pass

        field = fields.Function(
            serialize=lambda obj: obj.name.upper() + Context.get()["key"]
        )
        field.parent = Parent()
        with Context({"key": "BAR"}):
            assert field.serialize("key", user) == "MONTYBAR"

    def test_function_field_deserialization_with_context(self):
        class Parent(Schema):
            pass

        field = fields.Function(
            lambda x: None,
            deserialize=lambda val: val.upper() + Context.get()["key"],
        )
        field.parent = Parent()
        with Context({"key": "BAR"}):
            assert field.deserialize("foo") == "FOOBAR"

    def test_decorated_processors_with_context(self):
        NumDictContext = Context[dict[int, int]]

        class MySchema(Schema):
            f_1 = fields.Integer()
            f_2 = fields.Integer()
            f_3 = fields.Integer()
            f_4 = fields.Integer()

            @pre_dump
            def multiply_f_1(self, item, **kwargs):
                item["f_1"] *= NumDictContext.get()[1]
                return item

            @pre_load
            def multiply_f_2(self, data, **kwargs):
                data["f_2"] *= NumDictContext.get()[2]
                return data

            @post_dump
            def multiply_f_3(self, item, **kwargs):
                item["f_3"] *= NumDictContext.get()[3]
                return item

            @post_load
            def multiply_f_4(self, data, **kwargs):
                data["f_4"] *= NumDictContext.get()[4]
                return data

        schema = MySchema()

        with NumDictContext({1: 2, 2: 3, 3: 4, 4: 5}):
            assert schema.dump({"f_1": 1, "f_2": 1, "f_3": 1, "f_4": 1}) == {
                "f_1": 2,
                "f_2": 1,
                "f_3": 4,
                "f_4": 1,
            }
            assert schema.load({"f_1": 1, "f_2": 1, "f_3": 1, "f_4": 1}) == {
                "f_1": 1,
                "f_2": 3,
                "f_3": 1,
                "f_4": 5,
            }

    def test_validates_schema_with_context(self):
        class MySchema(Schema):
            f_1 = fields.Integer()
            f_2 = fields.Integer()

            @validates_schema
            def validate_schema(self, data, **kwargs):
                if data["f_2"] != data["f_1"] * Context.get():
                    raise ValidationError("Fail")

        schema = MySchema()

        with Context(2):
            schema.load({"f_1": 1, "f_2": 2})
            with pytest.raises(ValidationError) as excinfo:
                schema.load({"f_1": 1, "f_2": 3})
            assert excinfo.value.messages["_schema"] == ["Fail"]



import json

from marshmallow import EXCLUDE, Schema
from marshmallow.fields import Integer, String


# Test that valid `Meta` class attributes pass type checking
class MySchema(Schema):
    foo = String()
    bar = Integer()

    class Meta(Schema.Meta):
        fields = ("foo", "bar")
        additional = ("baz", "qux")
        include = {
            "foo2": String(),
        }
        exclude = ("bar", "baz")
        many = True
        dateformat = "%Y-%m-%d"
        datetimeformat = "%Y-%m-%dT%H:%M:%S"
        timeformat = "%H:%M:%S"
        render_module = json
        ordered = False
        index_errors = True
        load_only = ("foo", "bar")
        dump_only = ("baz", "qux")
        unknown = EXCLUDE
        register = False



"""Test utilities and fixtures."""

import datetime as dt
import functools
import typing
import uuid
from enum import Enum, IntEnum
from zoneinfo import ZoneInfo

import simplejson

from marshmallow import Schema, fields, missing, post_load, validate
from marshmallow.exceptions import ValidationError

central = ZoneInfo("America/Chicago")


class GenderEnum(IntEnum):
    male = 1
    female = 2
    non_binary = 3


class HairColorEnum(Enum):
    black = "black hair"
    brown = "brown hair"
    blond = "blond hair"
    red = "red hair"


class DateEnum(Enum):
    date_1 = dt.date(2004, 2, 29)
    date_2 = dt.date(2008, 2, 29)
    date_3 = dt.date(2012, 2, 29)


ALL_FIELDS = [
    fields.String,
    fields.Integer,
    fields.Boolean,
    fields.Float,
    fields.DateTime,
    fields.Time,
    fields.Date,
    fields.TimeDelta,
    fields.Dict,
    fields.Url,
    fields.Email,
    fields.UUID,
    fields.Decimal,
    fields.IP,
    fields.IPv4,
    fields.IPv6,
    fields.IPInterface,
    fields.IPv4Interface,
    fields.IPv6Interface,
    functools.partial(fields.Enum, GenderEnum),
    functools.partial(fields.Enum, HairColorEnum, by_value=fields.String),
    functools.partial(fields.Enum, GenderEnum, by_value=fields.Integer),
]


##### Custom asserts #####


def assert_date_equal(d1: dt.date, d2: dt.date) -> None:
    assert d1.year == d2.year
    assert d1.month == d2.month
    assert d1.day == d2.day


def assert_time_equal(t1: dt.time, t2: dt.time) -> None:
    assert t1.hour == t2.hour
    assert t1.minute == t2.minute
    assert t1.second == t2.second
    assert t1.microsecond == t2.microsecond


##### Validation #####


def predicate(
    func: typing.Callable[[typing.Any], bool],
) -> typing.Callable[[typing.Any], None]:
    def validate(value: typing.Any) -> None:
        if func(value) is False:
            raise ValidationError("Invalid value.")

    return validate


##### Models #####


class User:
    SPECIES = "Homo sapiens"

    def __init__(
        self,
        name,
        *,
        age=0,
        id_=None,
        homepage=None,
        email=None,
        registered=True,
        time_registered=None,
        birthdate=None,
        birthtime=None,
        balance=100,
        sex=GenderEnum.male,
        hair_color=HairColorEnum.black,
        employer=None,
        various_data=None,
    ):
        self.name = name
        self.age = age
        # A naive datetime
        self.created = dt.datetime(2013, 11, 10, 14, 20, 58)
        # A TZ-aware datetime
        self.updated = dt.datetime(2013, 11, 10, 14, 20, 58, tzinfo=central)
        self.id = id_
        self.homepage = homepage
        self.email = email
        self.balance = balance
        self.registered = registered
        self.hair_colors = list(HairColorEnum.__members__)
        self.sex_choices = list(GenderEnum.__members__)
        self.finger_count = 10
        self.uid = uuid.uuid1()
        self.time_registered = time_registered or dt.time(1, 23, 45, 6789)
        self.birthdate = birthdate or dt.date(2013, 1, 23)
        self.birthtime = birthtime or dt.time(0, 1, 2, 3333)
        self.activation_date = dt.date(2013, 12, 11)
        self.sex = sex
        self.hair_color = hair_color
        self.employer = employer
        self.relatives = []
        self.various_data = various_data or {
            "pets": ["cat", "dog"],
            "address": "1600 Pennsylvania Ave\nWashington, DC 20006",
        }

    @property
    def since_created(self):
        return dt.datetime(2013, 11, 24) - self.created

    def __repr__(self):
        return f"<User {self.name}>"


class Blog:
    def __init__(self, title, user, collaborators=None, categories=None, id_=None):
        self.title = title
        self.user = user
        self.collaborators = collaborators or []  # List/tuple of users
        self.categories = categories
        self.id = id_

    def __contains__(self, item):
        return item.name in [each.name for each in self.collaborators]


###### Schemas #####


class Uppercased(fields.String):
    """Custom field formatting example."""

    def _serialize(self, value, attr, obj, **kwargs):
        if value:
            return value.upper()
        return None


def get_lowername(obj):
    if obj is None:
        return missing
    if isinstance(obj, dict):
        return obj.get("name", "").lower()
    return obj.name.lower()


class UserSchema(Schema):
    name = fields.String()
    age: fields.Field = fields.Float()
    created = fields.DateTime()
    created_formatted = fields.DateTime(
        format="%Y-%m-%d", attribute="created", dump_only=True
    )
    created_iso = fields.DateTime(format="iso", attribute="created", dump_only=True)
    updated = fields.DateTime()
    species = fields.String(attribute="SPECIES")
    id = fields.String(dump_default="no-id")
    uppername = Uppercased(attribute="name", dump_only=True)
    homepage = fields.Url()
    email = fields.Email()
    balance = fields.Decimal()
    is_old: fields.Field = fields.Method("get_is_old")
    lowername = fields.Function(get_lowername)
    registered = fields.Boolean()
    hair_colors = fields.List(fields.Raw)
    sex_choices = fields.List(fields.Raw)
    finger_count = fields.Integer()
    uid = fields.UUID()
    time_registered = fields.Time()
    birthdate = fields.Date()
    birthtime = fields.Time()
    activation_date = fields.Date()
    since_created = fields.TimeDelta()
    sex = fields.Str(validate=validate.OneOf(list(GenderEnum.__members__)))
    various_data = fields.Dict()

    class Meta:
        render_module = simplejson

    def get_is_old(self, obj):
        if obj is None:
            return missing
        if isinstance(obj, dict):
            age = obj.get("age", 0)
        else:
            age = obj.age
        try:
            return age > 80
        except TypeError as te:
            raise ValidationError(str(te)) from te

    @post_load
    def make_user(self, data, **kwargs):
        return User(**data)


class UserExcludeSchema(UserSchema):
    class Meta:
        exclude = ("created", "updated")


class UserIntSchema(UserSchema):
    age = fields.Integer()


class UserFloatStringSchema(UserSchema):
    age = fields.Float(as_string=True)


class ExtendedUserSchema(UserSchema):
    is_old = fields.Boolean()


class UserRelativeUrlSchema(UserSchema):
    homepage = fields.Url(relative=True)


class BlogSchema(Schema):
    title = fields.String()
    user = fields.Nested(UserSchema)
    collaborators = fields.List(fields.Nested(UserSchema()))
    categories = fields.List(fields.String)
    id = fields.String()


class BlogOnlySchema(Schema):
    title = fields.String()
    user = fields.Nested(UserSchema)
    collaborators = fields.List(fields.Nested(UserSchema(only=("id",))))


class BlogSchemaExclude(BlogSchema):
    user = fields.Nested(UserSchema, exclude=("uppername", "species"))


class BlogSchemaOnlyExclude(BlogSchema):
    user = fields.Nested(UserSchema, only=("name",), exclude=("name", "species"))


class mockjson:  # noqa: N801
    @staticmethod
    def dumps(val):
        return b"{'foo': 42}"

    @staticmethod
    def loads(val):
        return {"foo": 42}
